/*
Copyright (c) 2014, Romain Dolbeau, unless otherwise noted.
No claims is made upon work by others.

For the work by Romain Dolbeau:
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

For the work by others:
See the indicated reference for the relevant license.
*/

#include <iostream>
#include <cstdio>
#include <cstdlib>
#if __cplusplus >= 201103L
#include <cstdint>
#else
#define uint64_t unsigned long long
#define uint32_t unsigned int
#define uint16_t unsigned short
#define uint8_t unsigned char
#endif

#include "chacha.h"

extern "C" {
unsigned int sleep(unsigned int seconds);
}


#define CHECK(K)                                                        \
  do {                                                                  \
    err = K;                                                            \
    if (err) {                                                          \
      fprintf(stderr, "Oups, "#K" failed with %d (%s)\n", err, cudaGetErrorString(err)); \
      fflush(stderr);                                                   \
      exit(-2);                                                         \
    } } while (0)

#define rotr(x,n)   (((x) >> ((int)(n))) | ((x) << (32 - (int)(n))))
#define rotl(x,n)   (((x) << ((int)(n))) | ((x) >> (32 - (int)(n))))
#if __CUDA_ARCH__ < 320
#define GROTATE(v,n) rotl(v,n)
#else
#define GROTATE(v,n) __funnelshift_rc(v,v,32-n)
#endif
#define GXOR(v,w) ((v) ^ (w))
#define GPLUS(v,w) ((v) + (w))
#define GPLUSONE(v) (GPLUS((v),1))
__device__ static inline uint32_t rotr8__gpu(const uint32_t x) {
  uint32_t tmp0;
  asm("prmt.b32  %0, %1, %1, %2;" : "=r"(tmp0) : "r"(x), "r"(0x0321));
  return tmp0;
}
__device__ static inline uint32_t rotr16__gpu(const uint32_t x) {
  uint32_t tmp0;
  asm("prmt.b32  %0, %1, %1, %2;" : "=r"(tmp0) : "r"(x), "r"(0x1032));
  return tmp0;
}
__device__ static inline uint32_t rotr24__gpu(const uint32_t x) {
  uint32_t tmp0;
  asm("prmt.b32  %0, %1, %1, %2;" : "=r"(tmp0) : "r"(x), "r"(0x2103));
  return tmp0;
}

#define QUARTERROUND(a,b,c,d)                             \
  x[a] = GPLUS(x[a],x[b]); x[d] = GROTATE(GXOR(x[d],x[a]),16); \
  x[c] = GPLUS(x[c],x[d]); x[b] = GROTATE(GXOR(x[b],x[c]),12); \
  x[a] = GPLUS(x[a],x[b]); x[d] = GROTATE(GXOR(x[d],x[a]), 8); \
  x[c] = GPLUS(x[c],x[d]); x[b] = GROTATE(GXOR(x[b],x[c]), 7)

#define QUARTERROUND_(a,b,c,d)                                   \
  x[a] = GPLUS(x[a],x[b]); x[d] = rotr16__gpu(GXOR(x[d],x[a])); \
  x[c] = GPLUS(x[c],x[d]); x[b] = GROTATE(GXOR(x[b],x[c]),12);  \
  x[a] = GPLUS(x[a],x[b]); x[d] = rotr24__gpu(GXOR(x[d],x[a])); \
  x[c] = GPLUS(x[c],x[d]); x[b] = GROTATE(GXOR(x[b],x[c]), 7)

__global__ void encrypt_bytes_gpu(uint32_t *input, const uint8_t *m, uint8_t *c, uint32_t bytes) {
  const uint32_t tx = threadIdx.x;
  const uint32_t b = (((blockIdx.x+blockIdx.y*gridDim.x)*blockDim.x) + threadIdx.x);
//   __shared__ uint32_t input[16];
  const uint32_t *mm = (uint32_t*)(m+64*b);
  uint32_t *mc = (uint32_t*)(c+64*b);
  uint32_t x[16];
  uint32_t i;
  
//   if (tx < 16)
//     input[tx] = gk[tx];
//   __syncthreads();
#define MAKE16(X)                                                       \
  X(0);X(1);X(2);X(3);X(4);X(5);X(6);X(7);X(8);X(9);X(10);X(11);X(12);X(13);X(14);X(15)
// #define MAKE16(X)                               \
//   for (i = 0 ; i < 16 ; i++) X(i)

#define COPY(i)                                 \
  x[i] = input[i]
  MAKE16(COPY);
  x[12] += b; /* fixme : x[13]*/

#if 1
  for (i = 20;i > 0;i -= 2) {
    QUARTERROUND( 0, 4, 8,12);
    QUARTERROUND( 1, 5, 9,13);
    QUARTERROUND( 2, 6,10,14);
    QUARTERROUND( 3, 7,11,15);
    QUARTERROUND( 0, 5,10,15);
    QUARTERROUND( 1, 6,11,12);
    QUARTERROUND( 2, 7, 8,13);
    QUARTERROUND( 3, 4, 9,14);
  }
#else
#define TWOROUNDS(X)                            \
  QUARTERROUND( 0, 4, 8,12);                    \
  QUARTERROUND( 1, 5, 9,13);                    \
  QUARTERROUND( 2, 6,10,14);                    \
  QUARTERROUND( 3, 7,11,15);                    \
  QUARTERROUND( 0, 5,10,15);                    \
  QUARTERROUND( 1, 6,11,12);                    \
  QUARTERROUND( 2, 7, 8,13);                    \
  QUARTERROUND( 3, 4, 9,14)
#define MAKE10(X)                                                       \
  X(0);X(1);X(2);X(3);X(4);X(5);X(6);X(7);X(8);X(9)
  MAKE10(TWOROUNDS);
#endif

#define ADD(i) \
  x[i] = GPLUS(x[i],input[i])
  MAKE16(ADD);
  x[12]+=b; /* fixme : x[13]*/
#define XOR(i) \
  mc[i] = mm[i] ^ x[i]
  MAKE16(XOR);
}

#define SIZE (8*1024*1024)
int main(int argc, char **argv) {
  uint8_t *in, *out0, *out1;
  uint8_t *gin, *gout;
  cudaError_t err;
  uint32_t key[8] = {0xc47b0294, 0xdbbbee0f, 0xec4757f2, 0x2ffeee35, 0x87ca4730, 0xc3d33b69, 0x1df38bab, 0x076bc558 };
  uint32_t iv[2] = {0xdeadbeef, 0xbadc0ffe};
  uint32_t keyinput[16];
  uint32_t *gkeyinput;
  uint32_t i, j;
  double t0, t1, t0b, t1b;
  uint32_t count;

  in = (uint8_t*)malloc((size_t)SIZE);
  out0 = (uint8_t*)malloc((size_t)SIZE);
  out1 = (uint8_t*)malloc((size_t)SIZE);
  
  CHECK(cudaMalloc((void**)&gin,(size_t)SIZE));
  CHECK(cudaMalloc((void**)&gout,(size_t)SIZE));
  CHECK(cudaMalloc((void**)&gkeyinput,(size_t)16*sizeof(uint32_t)));

  srandom(100);
  for (i = 0 ; i < SIZE/sizeof(uint32_t) ; i++) {
    ((uint32_t*)in)[i] = 0;//random();
  }

  keysetup(keyinput, (const unsigned char *)key, 256, 0);

#define ITER 3
  for (i = 0 ; i < ITER ;i++) {
    ivsetup(keyinput, (const unsigned char *)iv);
    t0 = wallclock();
    encrypt_bytes(keyinput, in, out0, SIZE);
    t1 = wallclock();
    printf("%zd B in %lf s = %lf MB/s\n",
           SIZE, t1-t0, ((double)SIZE)/(1000000. * (t1-t0)));
  }
  CHECK(cudaMemcpy(gin,in,SIZE,cudaMemcpyHostToDevice));
  for (i = 0 ; i < ITER ;i++) {
    ivsetup(keyinput, (const unsigned char *)iv);
    CHECK(cudaMemcpy(gkeyinput,keyinput,16*sizeof(uint32_t),cudaMemcpyHostToDevice));
    dim3 b, g;
    b.x=256;
    b.y=1;
    b.z=1;
    g.x = (SIZE/64)/256;
    g.y = 1;
    g.z = 1;
    t0 = wallclock();
    encrypt_bytes_gpu<<<g,b,0,0>>>(gkeyinput, gin, gout, SIZE);
    CHECK(cudaDeviceSynchronize());
    t1 = wallclock();
    printf("%zd B in %lf s = %lf MB/s\n",
           SIZE, t1-t0, ((double)SIZE)/(1000000. * (t1-t0)));
  }
  CHECK(cudaMemcpy(out1,gout,SIZE,cudaMemcpyDeviceToHost));
  count = 0;
  for (i = 0 ; i < SIZE && count < 64; i++) {
    if (out1[i] != out0[i]) {
      count ++;
      fprintf(stderr, "@%d: 0x%02x != 0x%02x\n", i, out0[i], out1[i]);
    }
  }
  
  CHECK(cudaFree(gin));
  CHECK(cudaFree(gout));
  CHECK(cudaFree(gkeyinput));
  
  
  free(in);
  free(out0);
  free(out1);
}
